//=============================================================================
// JPT_ChangeFollower
//=============================================================================

var Imported = Imported || {};
Imported['JPT_ChangeFollower'] = 1.00;

/*:ja
 * @plugindesc マップ隊列組み換え
 * @author Kawamura
 *
 * @param MaxFollower
 * @desc 同時に隊列表示できるアクターの最大数です。(最小値5)
 * @default 5
 *
 * @help
 *-------------------------------------------
 * ChangeFollower
 *-------------------------------------------
 * 指定したアクターIDで隊列を組みます
 * プレーヤーも変更されます
 *
 * ※マップ上の表示のみ変更します
 * ステータスやバトルには影響しません
 *
 * ChangeFollower 39 40
 * アクターID 39 40 の隊列にまります
 *
 *
 *-------------------------------------------
 * ReChangeFollower
 *-------------------------------------------
 * ChangeFollowerで変更した隊列を元の隊列に戻します
 *
 *
 *-------------------------------------------
 * AddNpcFollower
 *-------------------------------------------
 * 現在の隊列の任意の位置にNPCを追加します
 *
 * AddNpcFollower 9 1
 * アクター9を隊列1番の後ろに追加します
 *
 * AddNpcFollower 9 END
 * アクター9を現在の隊列の最後尾に追加します
 *
 * ※ChangeFollower中は機能しません
 *
 *
 *-------------------------------------------
 * RemoveNpcFollower
 *-------------------------------------------
 * 追加したNPCを隊列から削除します
 *
 * RemoveNpcFollower 1
 * 隊列1番の後ろのNPCを削除します
 *
 * RemoveNpcFollower END
 * NPCの中で最後尾のNPCを削除します
 *
 * RemoveNpcFollower ALL
 * 追加したNPCをすべて削除します
 *
 * ※ChangeFollower中は機能しません
 *
 */

/*
 *
 */

(function() {

	'use strict';


	var getParamNumber = function(value, min, max) {
        if (arguments.length < 2) min = -Infinity;
        if (arguments.length < 3) max = Infinity;
        return value.clamp(min, max);
    };

    var getMaxNpc = function(value){
    	return value - Game_Party.prototype.maxBattleMembers();
    }

	var parameters = PluginManager.parameters('JPT_ChangeFollower');

	var changeFollowerMaxNpc = getMaxNpc( getParamNumber( parseInt(parameters['MaxFollower'] || 5) , 5) );

	////////////////////////////////////////////////////////////////////////////////////
    // Game_Interpreter
	////////////////////////////////////////////////////////////////////////////////////

    var _JPT_ChangeFollower_Game_Interpreter_pluginCommand      = Game_Interpreter.prototype.pluginCommand;
    Game_Interpreter.prototype.pluginCommand = function(command, args) {
    	_JPT_ChangeFollower_Game_Interpreter_pluginCommand.call(this, command, args);

    	switch(command){
    	case 'ChangeFollower':
    		$gamePlayer.setViewActorId(args[0]);
    		$gameParty.followerActors = [];
    		for(var i=1; i<args.length;i++){
    			$gameParty.followerActors.push(args[i]);
    		}
    		$gameParty.pushNpc();
    		break;
    	case 'ReChangeFollower':
    		$gamePlayer.initViewActorId();
    		$gameParty.popNpc();
    		break;
    	case 'AddNpcFollower':
    		$gameParty.addNpc(args[0],args[1]);
    		break;
    	case 'RemoveNpcFollower':
    		$gameParty.removeNpc(args[0]);
    		break;
    	}
    };



	////////////////////////////////////////////////////////////////////////////////////
	// Game_Party
	////////////////////////////////////////////////////////////////////////////////////

	var _JPT_ChangeFollower_Game_Party_initialize = Game_Party.prototype.initialize;
	Game_Party.prototype.initialize = function() {
		_JPT_ChangeFollower_Game_Party_initialize.call(this, arguments);
		this.initNpc();
	};

	Game_Party.prototype.initNpc = function() {
        this._npcs       = [];
        this._npcIndexes = [];
        this.followerActors = [];
    };

    Game_Party.prototype.npcs = function() {
    	return this._npcs;
    }

    Game_Party.prototype.isNpcInvalid = function() {
        return !this._npcs;
    };

    Game_Party.prototype.initNpcIfNeed = function() {
        if (this.isNpcInvalid()) {
            this.initNpc();
            $gamePlayer.followers().initNpc();
        }
    };

    Game_Party.prototype.pushNpc = function() {

    	if( (changeFollowerMaxNpc + $gameParty.maxBattleMembers() - 1 ) >= this.followerActors.length){
    		//$gamePlayer.followers().pushNpc();
    		$gamePlayer.refresh();
    		$gameMap.requestRefresh();
    	}
    };

    Game_Party.prototype.popNpc = function() {
    	if(this.followerActors.length > 0){
	    	//$gamePlayer.followers().popNpc();
	    	this.followerActors = [];
	        $gamePlayer.refresh();
	        $gameMap.requestRefresh();
		}
    };

    Game_Party.prototype.addNpc = function(actorId, index) {
    	if(this.followerActors.length == 0 && changeFollowerMaxNpc > this._npcs.length ){
    		index = this.inprIndexString(index,true);
	    	this._npcs.push(actorId);
	    	this._npcIndexes.push(index);
	    	//$gamePlayer.followers().addNpc();
	    	$gamePlayer.refresh();
	    	$gameMap.requestRefresh();
    	}
    };


    Game_Party.prototype.inprIndexString = function(index,isAdd){
    	var indexNum = 1;
    	if( isNaN(index) ){
    		if(index.toUpperCase()==='END'){
    			if(isAdd){
    				indexNum = $gameParty.battleMembers().length + this._npcs.length;
    			}else{
    				indexNum = this._npcIndexes[0];
    				for(var i=0; i<this._npcIndexes.length; i++){
    					if(this._npcIndexes[i]>indexNum){
    						indexNum = this._npcIndexes[i];
    					}
    				}
    			}
    		}else if(index.toUpperCase()==='ALL' && !isAdd){
    			indexNum = 'ALL';
    		}

    	}else{
    		indexNum = parseInt(index, 10);
    	}
    	return indexNum;
    }

    Game_Party.prototype.removeNpc = function(index) {
    	if(this._npcs.length>0 && this.followerActors.length == 0){
    		index = this.inprIndexString(index,false);
    		if(index==='ALL'){
    			this.initNpc();
    		}else{
		        for (var i = 0, n = this._npcs.length; i < n; i++) {
		            if (this._npcIndexes[i] == index) {
		                this._npcs.splice(i, 1);
		                this._npcIndexes.splice(i, 1);
		                //i--;
		                break;
		            }
		        }
    		}
	        //$gamePlayer.followers().removeNpc();
	        $gamePlayer.refresh();
	        $gameMap.requestRefresh();
    	}
    };


    Game_Party.prototype.npcMembers = function() {
        return this._npcs.map(function(id) {
            return $gameActors.actor(id);
        });
    };

    Game_Party.prototype.followMembers = function() {
        return this.followerActors.map(function(id) {
            return $gameActors.actor(id);
        });
    };

    Game_Party.prototype.visibleMembers = function() {
        return this._visibleMembers;
    };

    Game_Party.prototype.makeVisibleMembers = function() {
        var battleMembers  = this.battleMembers();
        var visibleMembers = [];

        if(this.followerActors.length > 0){

        	var followMembers     = this.followMembers();
        	visibleMembers.push(battleMembers[0]);

        	for (var i = 0, n = this.followerActors.length + 1; i < n; i++) {
        		visibleMembers.push(followMembers[i]);
        	}

        }else{

        	var npcMembers = this.npcMembers();
        	for (var i = 0, n = this.maxBattleMembers() + 1 + npcMembers.length; i < n; i++) {
        		for (var j = 0, m = npcMembers.length; j < m; j++) {
	                if (this._npcIndexes[j] == i){
	                	visibleMembers.push(npcMembers[j]);
	                }
	            }
	            if (battleMembers.length > i){
	            	visibleMembers.push(battleMembers[i]);
	            }
        	}
        }

        this._visibleMembers = visibleMembers;
    };


    ////////////////////////////////////////////////////////////////////////////////////
	// Game_Followers
	////////////////////////////////////////////////////////////////////////////////////
    var _JPT_ChangeFollower_Game_Followers_actor = Game_Followers.prototype.initialize;
    Game_Followers.prototype.initialize = function() {
    	_JPT_ChangeFollower_Game_Followers_actor.call(this, arguments);
        this.initNpc();
    };


    Game_Followers.prototype.initNpc = function() {
        var memberLength = $gameParty.maxBattleMembers();
        for (var i = 0; i < changeFollowerMaxNpc; i++) {
            this._data.push(new Game_Follower(memberLength + i));
        }
    };

    /*
    Game_Followers.prototype.pushNpc = function() {
        var memberLength = $gameParty.maxBattleMembers();

        var len = memberLength + $gameParty.followerActors.length + $gameParty.npcs().length - this._data.length;

        if(len <= 0)return;

        for (var i = 0; i < len; i++) {
            this._data.push(new Game_Follower(memberLength + i));
        }
    };

    Game_Followers.prototype.popNpc = function() {
        var memberLength = $gameParty.maxBattleMembers();
        for (var i = 0; i < $gameParty.followerActors.length; i++) {
            this._data.pop();
        }
    };

    Game_Followers.prototype.addNpc = function() {
    	if( $gameParty.npcs().length <= 0 )return;
    	var memberLength = $gameParty.maxBattleMembers() + $gameParty.npcs().length - 1;
    	this._data.push(new Game_Follower(memberLength));
    }

    Game_Followers.prototype.removeNpc = function() {
        var len = $gameParty.maxBattleMembers() + $gameParty.npcs().length - this._data.length;

        if(len >= 0)return;

    	this._data.pop();
    }
    */

    ////////////////////////////////////////////////////////////////////////////////////
	// Game_Follower
	////////////////////////////////////////////////////////////////////////////////////

    var _JPT_ChangeFollower_Game_Follower_actor = Game_Follower.prototype.actor;
    Game_Follower.prototype.actor = function() {
    	_JPT_ChangeFollower_Game_Follower_actor.call(this);
        return $gameParty.visibleMembers()[this._memberIndex];
    };


    ////////////////////////////////////////////////////////////////////////////////////
	// Game_Player
	////////////////////////////////////////////////////////////////////////////////////

    var _JPT_ChangeFollower_Game_Player_refresh = Game_Player.prototype.refresh;
    Game_Player.prototype.refresh = function() {
        $gameParty.makeVisibleMembers();
        _JPT_ChangeFollower_Game_Player_refresh.call(this);

        if(this.viewActorId() > 0){
        	var actor = this.viewActor();
            var characterName = actor ? actor.characterName() : '';
            var characterIndex = actor ? actor.characterIndex() : 0;
            this.setImage(characterName, characterIndex);
            this._followers.refresh();
        }
    };

    var _JPT_ChangeFollower_Game_Player_initMembers = Game_Player.prototype.initMembers;
    Game_Player.prototype.initMembers = function() {
    	_JPT_ChangeFollower_Game_Player_initMembers.call(this);
    	this.initViewActorId();
    };

    Game_Player.prototype.initViewActorId = function(){
    	this._viewActorId = 0;
    }

    Game_Player.prototype.viewActor = function(){
    	return $gameActors.actor(this._viewActorId);
    }

    Game_Player.prototype.viewActorId = function(){
    	return this._viewActorId;
    }

    Game_Player.prototype.setViewActorId = function(actorId){
    	this._viewActorId = actorId;
    }

}());
